import six

from ipalib import api, errors
from ipalib import Bool, Str, errors, Int, Flag
from ipalib import output
from ipalib.plugable import Registry
from ipaserver.plugins.baseldap import (
    LDAPObject,
    LDAPCreate,
    LDAPRetrieve,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPAddMember,
    LDAPRemoveMember)
from ipalib import _, ngettext
from ipapython.dn import DN
from ipaserver.plugins.group import group, group_add, group_mod
from ipaserver.plugins.host import host, host_add
from ipaserver.plugins.user import user, user_add, user_mod, user_show

if six.PY3:
    unicode = str

__doc__ = _("""
Mail alias

EXAMPLES:

 Add a new mail alias:
   ipa mailalias-add --desc='3rd floor lab' myalias myaliastarget

 Delete a mail alias:
   ipa mailalias-del myalias

 Modify information about a mail alias:
   ipa mailalias-mod --desc='3rd floor lab' myalias

 Show information about a mail alias:
   ipa mailalias-show myalias

""")

register = Registry()

@register()
class mailalias(LDAPObject):
    """
    Mail alias object
    """
    container_dn = DN(('cn', 'alias'), ('cn', 'mail'))
    object_name = _('mail alias')
    object_name_plural = _('mail aliases')
    object_class = ['imsmailaliasmap' ]
    permission_filter_objectclasses = ["imsmailaliasmap"]
    search_attributes = [
        'imsmailaliassource', 'imsmailaliastargets', 'description'
    ]
    default_attributes = [
        'imsmailaliassource', 'imsmailaliastargets', 'description'
    ]
    rdn_attribute = 'imsmailaliassource'
    allow_rename = True
    bindable = False

    managed_permissions = {
        'System: Read Mail Alias': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'imsmailaliassource', 'imsmailaliastargets',
                'description', 'objectclass',
            },
        },
        'System: Add Mail Alias': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///imsmailaliassource=*,cn=alias,cn=mail,$SUFFIX")(version 3.0;acl "permission:Add Mail Alias";allow (add) groupdn = "ldap:///cn=Add Mail Alias,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mail Alias Administrators'},
        },
        'System: Modify Mail Alias': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'imsmailaliassource', 'imsmailaliastargets', 'description'
            },
            'replaces': [
                '(targetattr = "imsmailaliassource || imsmailaliastargets || description" )(target = "ldap:///imsmailalias=*,cn=alias,cn=mail,$SUFFIX")(version 3.0;acl "permission:Modify Mail Alias";allow (write) groupdn = "ldap:///cn=Modify Mail Alias,cn=permissions,cn=pbac,$SUFFIX";)',

            ],
            'default_privileges': {'Mail Alias Administrators'},
        },
        'System: Remove Mail Alias': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///imsmailaliassource=*,cn=alias,cn=mail,$SUFFIX")(version 3.0;acl "permission:Remove Mail Alias";allow (delete) groupdn = "ldap:///cn=Remove Mail Alias,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mail Alias Administrators'},
        },
    }

    label = _('Mail Aliases map')
    label_singular = _('Mail map alias')

    takes_params = (
        Str('imsmailaliassource',
            cli_name='alias',
            label=_('Mail map alias source'),
            primary_key=True
        ),
        Str('imsmailaliastargets+',
            cli_name='target',
            label=_('Mail map alias targets')
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('A description of this mail alias'),
        ),
    )

@register()
class mailalias_add(LDAPCreate):
    __doc__ = _('Add a new mail map alias.')

    has_output_params = LDAPCreate.has_output_params
    msg_summary = _('Added mail map alias "%(value)s"')
    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        assert isinstance(dn, DN)
        try:
            # check duplicity with hostgroups first to provide proper error
            api.Object['mailalias'].get_dn_if_exists(keys[-1])
            self.obj.handle_duplicate_entry(*keys)
        except errors.NotFound:
            pass

        return dn

@register()
class mailalias_del(LDAPDelete):
    __doc__ = _('Delete a mail map alias.')

    msg_summary = _('Deleted mail map alias "%(value)s"')

@register()
class mailalias_mod(LDAPUpdate):
    __doc__ = _('Modify information about a mail map alias.')

    has_output_params = LDAPUpdate.has_output_params

    msg_summary = _('Modified mail map alias "%(value)s"')

@register()
class mailalias_find(LDAPSearch):
    __doc__ = _('Search for mail map aliases.')

    has_output_params = LDAPSearch.has_output_params

    msg_summary = ngettext(
        '%(count)d mail map alias matched', '%(count)d mail map aliass matched', 0
    )

@register()
class mailalias_show(LDAPRetrieve):
    __doc__ = _('Display information about a mail map alias.')
    has_output_params = LDAPRetrieve.has_output_params

