import six

from ipalib import api, errors
from ipalib import Bool, Str, errors, Int, Flag, DNParam
from ipalib import output
from ipalib.plugable import Registry
from ipaserver.plugins.baseldap import (
    LDAPObject,
    LDAPCreate,
    LDAPRetrieve,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPAddMember,
    LDAPRemoveMember)
from ipalib import _, ngettext
from ipapython.dn import DN
from ipaserver.plugins.group import group, group_add, group_mod
from ipaserver.plugins.host import host, host_add
from ipaserver.plugins.user import user, user_add, user_mod, user_show

if six.PY3:
    unicode = str

__doc__ = _("""
Mailbox

System mailbox

EXAMPLES:

 Add a new mailbox:
   ipa mailbox-add --desc='3rd floor lab' mymailbox

 Delete a mailbox:
   ipa mailbox-del mymailbox

 Modify information about a mailbox:
   ipa mailbox-mod --desc='3rd floor lab' mymailbox

 Manage mailbox membership, commands:
   ipa mailbox-add-member --machine=test mymailbox

 Manage mailbox membership, commands:
   ipa mailbox-remove-member --machine=test mymailbox

 Show information about a mailbox:
   ipa mailbox-show mymailbox

""")

register = Registry()

mailbox_output_params = (
    Str('imsmailboxmember_user?',
        label=_('Member users'),
    ),
    Str('imsmailboxmember_group?',
        label=_('Member groups'),
    ),
)

@register()
class mailbox(LDAPObject):
    """
    Mailbox object
    """
    container_dn = DN(('cn', 'mailbox'), ('cn', 'mail'))
    object_name = _('mailbox')
    object_name_plural = _('mailboxes')
    object_class = ['imsmailboxsystem' ]
    permission_filter_objectclasses = ['imsmailboxsystem']
    search_attributes = [
        'imsmailaddress', 'description'
    ]
    default_attributes = [
        'imsmailaddress', 'description', 'imsmailboxmember',
        'imsmailboxquota', 'imsmailboxclassdn', 'imsmailboxtransport'
    ]
    rdn_attribute = 'imsmailaddress'
    allow_rename = True
    attribute_members = {
        'imsmailboxmember': ['user', 'group']
    }
    bindable = False

    managed_permissions = {
        'System: Read Mailboxes': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'imsmailaddress', 'description',
                'imsmailboxclassdn', 'imsmailboxquota', 'imsmailboxtransport',
                'objectclass'
            },
        },
        'System: Read Mailboxes Membership': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'imsmailboxmember'
            },
        },
        'System: Add Mailboxes': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///imsmailaddress=*,cn=mailbox,cn=mail,$SUFFIX")(version 3.0;acl "permission:Add Mailboxes";allow (add) groupdn = "ldap:///cn=Add Mailboxes,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mailboxes Administrators'},
        },
        'System: Modify Mailboxes Membership': {
            'ipapermright': {'write'},
            'ipapermtargetfilter': [
                '(objectclass=imsmailboxsystem)',
            ],
            'ipapermdefaultattr': {'imsmailboxmember'},
            'replaces': [
                '(targetattr = "imsmailboxmember")(target = "ldap:///imsmailaddress=*,cn=hostgroups,cn=accounts,$SUFFIX")(version 3.0;acl "permission:Modify Mailboxes membership";allow (write) groupdn = "ldap:///cn=Modify Mailboxes  membership,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mailboxes Administrators'},
        },
        'System: Modify Mailboxes': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'imsmailaddress', 'description',
                'imsmailboxclassdn', 'imsmailboxquota', 'imsmailboxtransport',
            },
            'replaces': [
                '(targetattr = "cn || description || imsmailboxclassdn || imsmailboxquota || imsmailboxtransport" )(target = "ldap:///imsmailaddress=*,cn=mailbox,cn=mail,$SUFFIX")(version 3.0;acl "permission:Modify Mailboxes";allow (write) groupdn = "ldap:///cn=Modify Mailboxes,cn=permissions,cn=pbac,$SUFFIX";)',

            ],
            'default_privileges': {'Mailboxes Administrators'},
        },
        'System: Remove Mailboxes': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///imsmailaddress=*,cn=mailbox,cn=mail,$SUFFIX")(version 3.0;acl "permission:Remove Mailboxes";allow (delete) groupdn = "ldap:///cn=Remove Mailboxes,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mailboxes Administrators'},
        },
    }

    label = _('Mailboxes')
    label_singular = _('Mailbox')

    takes_params = (
        Str('imsmailaddress',
            cli_name='name',
            label=_('Mailbox address'),
            primary_key=True,
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('A description of this mailbox'),
        ),
        DNParam('imsmailboxclassdn?',
            cli_name='mailbox_class',
            label=_('Mailbox class')
        ),
        Int('imsmailboxquota?',
            cli_name='quota',
            label=_('Mailbox quota [MB]')
        ),
        Str('imsmailboxtransport?',
            cli_name='transport',
            label=_('Mailbox transport'),
            doc=_('Mailbox transport in transport:nexthop form eg. smtp:myhostname, lmtp:[192.168.1.1]')
        ),
    )


@register()
class mailbox_add(LDAPCreate):
    __doc__ = _('Add a new mailbox.')

    has_output_params = LDAPCreate.has_output_params + mailbox_output_params
    msg_summary = _('Added mailbox "%(value)s"')

@register()
class mailbox_del(LDAPDelete):
    __doc__ = _('Delete a mailbox.')

    msg_summary = _('Deleted mailbox "%(value)s"')

@register()
class mailbox_mod(LDAPUpdate):
    __doc__ = _('Modify information about a mailbox.')

    msg_summary = _('Modified mailbox "%(value)s"')

    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        if 'imsmailboxquota' in entry_attrs:
            try:
                quota = int(entry_attrs['imsmailboxquota'])
            except ValueError:
                raise errors.ValidationError(name='imsmailboxquota', error='not a number')
            quota_str = '*:storage={}M'.format(quota)
            entry_attrs['imsmailboxquota'] = quota_str

        return dn

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        if 'imsmailboxquota' in entry_attrs:
            try:
                _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxquota'][0]))
                entry_attrs['imsmailboxquota'] = str(limit)
            except ValueError:
                pass

        return dn

@register()
class mailbox_find(LDAPSearch):
    __doc__ = _('Search for mailboxes.')

    has_output_params = LDAPSearch.has_output_params + mailbox_output_params

    msg_summary = ngettext(
        '%(count)d mailbox matched', '%(count)d mailboxes matched', 0
    )

@register()
class mailbox_show(LDAPRetrieve):
    __doc__ = _('Display information about a mailbox.')
    has_output_params = LDAPRetrieve.has_output_params + mailbox_output_params

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        if 'imsmailboxquota' in entry_attrs:
            try:
                _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxquota'][0]))
                entry_attrs['imsmailboxquota'] = str(limit)
            except ValueError:
                pass

        return dn

@register()
class mailbox_add_member(LDAPAddMember):
    __doc__ = _('Add members to mailbox.')
    member_attributes = ['imsmailboxmember']

@register()
class mailbox_remove_member(LDAPRemoveMember):
    __doc__ = _('Remove members from mailbox.')
    member_attributes = ['imsmailboxmember']
