import six

from ipalib import api, errors
from ipalib import Bool, Str, errors, Int, Flag
from ipalib import output
from ipalib.plugable import Registry
from ipaserver.plugins.baseldap import (
    LDAPObject,
    LDAPCreate,
    LDAPRetrieve,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPAddMember,
    LDAPRemoveMember)
from ipalib import _, ngettext
from ipapython.dn import DN
from ipaserver.plugins.group import group, group_add, group_mod
from ipaserver.plugins.host import host, host_add
from ipaserver.plugins.user import user, user_add, user_mod, user_show

if six.PY3:
    unicode = str

__doc__ = _("""
Mailbox class

EXAMPLES:

 Add a new mailbox class:
   ipa mailboxclass-add --desc='3rd floor lab' myclass

 Delete a mailbox class:
   ipa mailboxclass-del myclass

 Modify information about a mailbox class:
   ipa mailboxclass-mod --desc='3rd floor lab' myclass

 Show information about a mailbox class:
   ipa mailboxclass-show myclass

""")

register = Registry()

def parse_quota_rule(quota_str):
    def norm_to_mb(limit_str):
        suffix = limit_str[-1]
        _limit = int(limit_str[0:-1])

        in_mb = 0
        if suffix == 'b':
            in_mb = _limit // 1024 // 1024
        elif suffix == 'k':
            in_mb = _limit // 1024
        elif suffix == 'M':
            in_mb = _limit
        elif suffix == 'G':
            in_mb = _limit * 1024
        elif suffix == 'T':
            in_mb = _limit * 1024 * 1024

        return in_mb

    parts = quota_str.split(':')

    mailbox = parts[0]
    limit_conf = parts[1].split('=')
    limit = None
    if len(limit_conf) > 1:
        limit_type = limit_conf[0]
        limit = limit_conf[1]

        if limit_type in ('storage', 'bytes'):
            limit = norm_to_mb(limit)
    else:
        limit_type = limit_conf[0]

    return mailbox, limit_type, limit



@register()
class mailboxclass(LDAPObject):
    """
    Mailbox class object
    """
    container_dn = DN(('cn', 'class'), ('cn', 'mail'))
    object_name = _('mailbox class')
    object_name_plural = _('mailbox classes')
    object_class = ['imsMailboxClass' ]
    permission_filter_objectclasses = ['imsMailboxClass']
    search_attributes = [
        'cn', 'description'
    ]
    default_attributes = [
        'cn', 'imsmailboxquota', 'description'
    ]
    rdn_attribute = 'cn'
    allow_rename = True
    bindable = False

    managed_permissions = {
        'System: Read Mailbox Class': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'cn', 'description', 'imsmailboxquota',
                'objectclass'
            },
        },
        'System: Add Mailbox Class': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///cn=*,cn=class,cn=mail,$SUFFIX")(version 3.0;acl "permission:Add Mailbox Class";allow (add) groupdn = "ldap:///cn=Add Mailbox Class,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mailbox Class Administrators'},
        },
        'System: Modify Mailbox Class': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'cn', 'description', 'imsmailboxquota'
            },
            'replaces': [
                '(targetattr = "cn || description || imsmailboxquota" )(target = "ldap:///cn=*,cn=class,cn=mail,$SUFFIX")(version 3.0;acl "permission:Modify Mailbox Class";allow (write) groupdn = "ldap:///cn=Modify Mailbox Class,cn=permissions,cn=pbac,$SUFFIX";)',

            ],
            'default_privileges': {'Mailbox Class Administrators'},
        },
        'System: Remove Mailbox Class': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///cn=*,cn=class,cn=mail,$SUFFIX")(version 3.0;acl "permission:Remove Mailbox Class";allow (delete) groupdn = "ldap:///cn=Remove Mailbox Class,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mailbox Class Administrators'},
        },
    }

    label = _('Mailbox Classes')
    label_singular = _('Mailbox class')

    takes_params = (
        Str('cn',
            cli_name='name',
            label=_('Mailbox class'),
            primary_key=True
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('A description of this mailbox class'),
        ),
        Int('imsmailboxquota?',
            cli_name='quota',
            label='Mailbox quota [MB]'
        ),
    )

@register()
class mailboxclass_add(LDAPCreate):
    __doc__ = _('Add a new mailbox class.')

    msg_summary = _('Added mailbox class "%(value)s"')

@register()
class mailboxclass_del(LDAPDelete):
    __doc__ = _('Delete a mailbox class.')

    msg_summary = _('Deleted mailbox class "%(value)s"')

@register()
class mailboxclass_mod(LDAPUpdate):
    __doc__ = _('Modify information about a mailbox class.')

    msg_summary = _('Modified mailbox class "%(value)s"')

    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        if 'imsmailboxquota' in entry_attrs:
            try:
                quota = int(entry_attrs['imsmailboxquota'])
            except ValueError:
                raise errors.ValidationError(name='imsmailboxquota', error='not a number')
            quota_str = '*:storage={}M'.format(quota)
            entry_attrs['imsmailboxquota'] = quota_str

        return dn

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        if 'imsmailboxquota' in entry_attrs:
            try:
                _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxquota'][0]))
                entry_attrs['imsmailboxquota'] = str(limit)
            except ValueError:
                pass

        return dn


@register()
class mailboxclass_find(LDAPSearch):
    __doc__ = _('Search for mailbox classs.')

    msg_summary = ngettext(
        '%(count)d mailbox class matched', '%(count)d mailbox classs matched', 0
    )

@register()
class mailboxclass_show(LDAPRetrieve):
    __doc__ = _('Display information about a mailbox class.')

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        if 'imsmailboxquota' in entry_attrs:
            try:
                _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxquota'][0]))
                entry_attrs['imsmailboxquota'] = str(limit)
            except ValueError:
                pass

        return dn

