import six

from ipalib import api, errors
from ipalib import Bool, Str, errors, Int, Flag, DNParam
from ipalib import output
from ipalib.plugable import Registry
from ipaserver.plugins.baseldap import (
    LDAPObject,
    LDAPCreate,
    LDAPRetrieve,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPAddMember,
    LDAPRemoveMember)
from ipalib import _, ngettext
from ipapython.dn import DN
from ipaserver.plugins.group import group, group_add, group_mod
from ipaserver.plugins.host import host, host_add
from ipaserver.plugins.user import user, user_add, user_mod, user_show

if six.PY3:
    unicode = str

__doc__ = _("""
Mail domain

EXAMPLES:

 Add a new mail domain:
   ipa maildomain-add --desc='3rd floor lab' mydomain

 Delete a mail domain:
   ipa maildomain-del mydomain

 Modify information about a mail domain:
   ipa maildomain-mod --desc='3rd floor lab' mydomain

 Show information about a mail domain:
   ipa maildomain-show mydomain

""")

register = Registry()

def parse_quota_rule(quota_str):
    def norm_to_mb(limit_str):
        suffix = limit_str[-1]
        _limit = int(limit_str[0:-1])

        in_mb = 0
        if suffix == 'b':
            in_mb = _limit // 1024 // 1024
        elif suffix == 'k':
            in_mb = _limit // 1024
        elif suffix == 'M':
            in_mb = _limit
        elif suffix == 'G':
            in_mb = _limit * 1024
        elif suffix == 'T':
            in_mb = _limit * 1024 * 1024

        return in_mb

    parts = quota_str.split(':')

    mailbox = parts[0]
    limit_conf = parts[1].split('=')
    limit = None
    if len(limit_conf) > 1:
        limit_type = limit_conf[0]
        limit = limit_conf[1]

        if limit_type in ('storage', 'bytes'):
            limit = norm_to_mb(limit)
    else:
        limit_type = limit_conf[0]

    return mailbox, limit_type, limit


@register()
class maildomain(LDAPObject):
    """
    Mail domain object
    """
    container_dn = DN(('cn', 'domain'), ('cn', 'mail'))
    object_name = _('mail domain')
    object_name_plural = _('mail domains')
    object_class = ['imsMailDomainObject' ]
    permission_filter_objectclasses = ["imsMailDomainObject"]
    search_attributes = [
        'imsmaildomain', 'description'
    ]
    default_attributes = [
        'imsmaildomain', 'imsmailboxdefaultclassdn', 'imsmailboxdefaultquota',
        'imsmailboxtransport', 'description'
    ]
    rdn_attribute = 'imsmaildomain'
    allow_rename = True
    bindable = False

    managed_permissions = {
        'System: Read Mail Domain': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'imsmaildomain', 'description',
                'imsmailboxdefaultclassdn', 'imsmailboxdefaultquota',
                'imsmailboxtransport', 'objectclass'
            },
        },
        'System: Add Mail Domain': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///imsmaildomain=*,cn=domain,cn=mail,$SUFFIX")(version 3.0;acl "permission:Add Mail Domain";allow (add) groupdn = "ldap:///cn=Add Mail Domain,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mail Domain Administrators'},
        },
        'System: Modify Mail Domain': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'imsmaildomain', 'description',
                'imsmailboxdefaultclassdn', 'imsmailboxdefaultquota',
                'imsmailboxtransport'
            },
            'replaces': [
                '(targetattr = "imsmaildomain || description || imsmailboxdefaultclassdn || imsmailboxdefaultquota || imsmailboxtransport" )(target = "ldap:///imsmaildomain=*,cn=domain,cn=mail,$SUFFIX")(version 3.0;acl "permission:Modify Mail Domain";allow (write) groupdn = "ldap:///cn=Modify Mail Domain,cn=permissions,cn=pbac,$SUFFIX";)',

            ],
            'default_privileges': {'Mail Domain Administrators'},
        },
        'System: Remove Mail Domain': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///imsmaildomain=*,cn=domain,cn=mail,$SUFFIX")(version 3.0;acl "permission:Remove Mail Domain";allow (delete) groupdn = "ldap:///cn=Remove Mail Domain,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Mail Domain Administrators'},
        },
    }

    label = _('Mail Domains')
    label_singular = _('Mail domain')

    takes_params = (
        Str('imsmaildomain',
            cli_name='name',
            label=_('Mail domain'),
            primary_key=True
        ),
        DNParam('imsmailboxdefaultclassdn?',
            cli_name='defaultclass',
            label=_('Mail domain default class')
        ),
        Int('imsmailboxdefaultquota?',
            cli_name='defaultquota',
            label=_('Mail domain default quota [MB]')
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('A description of this mail domain'),
        ),
        Str('imsmailboxtransport?',
            cli_name='transport',
            label=_('Mail transport'),
            doc=_('Mail transport in transport:nexthop form eg. smtp:myhostname, lmtp:[192.168.1.1]'),
        ),
    )

@register()
class maildomain_add(LDAPCreate):
    __doc__ = _('Add a new mail domain.')

    has_output_params = LDAPCreate.has_output_params
    msg_summary = _('Added mail domain "%(value)s"')
    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        assert isinstance(dn, DN)
        try:
            # check duplicity with hostgroups first to provide proper error
            api.Object['maildomain'].get_dn_if_exists(keys[-1])
            self.obj.handle_duplicate_entry(*keys)
        except errors.NotFound:
            pass

        return dn

@register()
class maildomain_del(LDAPDelete):
    __doc__ = _('Delete a mail domain.')

    msg_summary = _('Deleted mail domain "%(value)s"')

@register()
class maildomain_mod(LDAPUpdate):
    __doc__ = _('Modify information about a mail domain.')

    msg_summary = _('Modified mail domain "%(value)s"')

    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        if 'imsmailboxdefaultquota' in entry_attrs:
            try:
                quota = int(entry_attrs['imsmailboxdefaultquota'])
            except ValueError:
                raise errors.ValidationError(name='imsmailboxdefaultquota', error='not a number')
            quota_str = '*:storage={}M'.format(quota)
            entry_attrs['imsmailboxdefaultquota'] = quota_str

        return dn

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        if 'imsmailboxdefaultquota' in entry_attrs:
            try:
                _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxdefaultquota'][0]))
                entry_attrs['imsmailboxdefaultquota'] = str(limit)
            except ValueError:
                pass

        return dn


@register()
class maildomain_find(LDAPSearch):
    __doc__ = _('Search for mail domains.')

    msg_summary = ngettext(
        '%(count)d mail domain matched', '%(count)d mail domains matched', 0
    )

@register()
class maildomain_show(LDAPRetrieve):
    __doc__ = _('Display information about a mail domain.')

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        if 'imsmailboxdefaultquota' in entry_attrs:
            try:
                _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxdefaultquota'][0]))
                entry_attrs['imsmailboxdefaultquota'] = str(limit)
            except ValueError:
                pass

        return dn
