from ipalib import _, Bool, Str, errors, Int, Flag, DNParam
from ipalib import output
from ipalib.plugable import Registry
from ipalib.util import validate_domain_name
from ipapython.dn import DN
from ipapython.ipavalidate import Email
from ipaserver.plugins.baseldap import LDAPObject, LDAPUpdate, LDAPRetrieve, add_missing_object_class, LDAPQuery, \
    pkey_to_value
from ipaserver.plugins.group import group, group_add, group_mod
from ipaserver.plugins.host import host, host_add
from ipaserver.plugins.user import user, user_add, user_mod, user_show

import traceback
import logging

logger = logging.getLogger(__name__)

__doc__ = _("""
Mail server configuration

Configure mailboxes and mail routing.
""")

register = Registry()


class MailAlreadyMigratedError(errors.GenericError):
    format = _("Mail attributes for this entry were already migrated.")


def normalize_and_validate_email(email, config):
    # check if default email domain should be added
    defaultdomain = config.get('ipadefaultemaildomain', [None])[0]
    if email:
        norm_email = []
        if not isinstance(email, (list, tuple)):
            email = [email]
        for m in email:
            if isinstance(m, str):
                if '@' not in m and defaultdomain:
                    m = m + u'@' + defaultdomain
                if not Email(m):
                    raise errors.ValidationError(name='email',
                                                 error=_('invalid e-mail format: %(email)s') % dict(email=m))
                norm_email.append(m)
            else:
                if not Email(m):
                    raise errors.ValidationError(name='email',
                                                 error=_('invalid e-mail format: %(email)s') % dict(email=m))
                norm_email.append(m)
        return norm_email

    return email


def parse_quota_rule(quota_str):
    def norm_to_mb(limit_str):
        suffix = limit_str[-1]
        _limit = int(limit_str[0:-1])

        in_mb = 0
        if suffix == 'b':
            in_mb = _limit // 1024 // 1024
        elif suffix == 'k':
            in_mb = _limit // 1024
        elif suffix == 'M':
            in_mb = _limit
        elif suffix == 'G':
            in_mb = _limit * 1024
        elif suffix == 'T':
            in_mb = _limit * 1024 * 1024

        return in_mb

    parts = quota_str.split(':')

    mailbox = parts[0]
    limit_conf = parts[1].split('=')
    limit = None
    if len(limit_conf) > 1:
        limit_type = limit_conf[0]
        limit = limit_conf[1]

        if limit_type in ('storage', 'bytes'):
            limit = norm_to_mb(limit)
    else:
        limit_type = limit_conf[0]

    return mailbox, limit_type, limit


def normalize_mail_attrs(entry_attrs):
    if 'mail' in entry_attrs:
        if len(entry_attrs['mail']) > 1:
            raise errors.ValidationError(name='mail', error=_('Mail attribute has to be single-valued.'
                                                              'Use alias/imsmailsenderalias to specify more addresses.'))
        else:
            if 'imsmailaddress' in entry_attrs:
                entry_attrs['mail'] = [entry_attrs['imsmailaddress']]
            else:
                entry_attrs['imsmailaddress'] = entry_attrs['mail'][0]

    if 'imsmailaddress' in entry_attrs:
        entry_attrs['mail'] = [entry_attrs['imsmailaddress']]


user.takes_params += (
    Str('imsmailaddress?',
        cli_name='primary_mail',
        label=_('Primary mail address')
        ),
    Str('imsmailalias*',
        cli_name='alias',
        label=_('Mail aliases')
        ),
    Str('imsmailsenderalias*',
        cli_name='send_alias',
        label=_('Allowed sender aliases')
        ),
    Bool('imsmailreceiveexternally?',
         cli_name='can_receive_externally',
         label='Can receive external mails'),
    Bool('imsmailsendexternally?',
         cli_name='can_send_externally',
         label='Can send mails to external locations'),
    Int('imsmailboxquota?',
        cli_name='mailbox_quota',
        label='Mailbox quota [MB]'),
    DNParam('imsmailboxclassdn?',
        cli_name='mailbox_class',
        label=_('Mailbox class')
    )
)

user.default_attributes = user.default_attributes + ['imsmailalias', 'imsmailsenderalias', 'imsmailboxquota', 'imsmailboxclassdn']

user.managed_permissions = {**user.managed_permissions, **{
    'System: Read User Mail Attributes': {
        'ipapermbindruletype': 'all',
        'ipapermright': {'read', 'search', 'compare'},
        'ipapermdefaultattr': {
            'imsmailaddress', 'imsmailalias', 'imsmailsenderalias', 'imsmailreceiveexternally', 'imsmailsendexternally', 'imsmailboxquota',
            'imsmailboxclassdn'
        }
    },
    'System: Modify User Mail Attributes': {
        'ipapermbindruletype': 'permission',
        'ipapermright': {'write', 'add', 'delete'},
        'ipapermdefaultattr': {
            'imsmailaddress', 'imsmailalias', 'imsmailsenderalias', 'imsmailreceiveexternally', 'imsmailsendexternally', 'imsmailboxquota',
            'imsmailboxclassdn'
        }
    }
}}


def usershow_post_callback(self, ldap, dn, entry_attrs, *keys, **options):
    if 'imsmailboxquota' in entry_attrs:
        try:
            _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxquota'][0]))
            entry_attrs['imsmailboxquota'] = str(limit)
        except ValueError:
            pass

    return dn


user_show.register_post_callback(usershow_post_callback)


def useradd_pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
    add_missing_object_class(ldap, 'imsmailsender', dn, entry_attrs, update=False)
    add_missing_object_class(ldap, 'imsmailrecipent', dn, entry_attrs, update=False)
    add_missing_object_class(ldap, 'imsmailboxuser', dn, entry_attrs, update=False)

    if 'imsmailaddress' not in entry_attrs:
        if 'mail' in entry_attrs:
            entry_attrs['imsmailaddress'] = entry_attrs['mail'][0]

    normalize_mail_attrs(entry_attrs)

    mail_domain = entry_attrs['imsmailaddress'] .split("@")[1]

    if 'imsmailboxclassdn' not in entry_attrs:
        try:
            mail_domain_config = self.api.Command['maildomain_show'](all=True, raw=True, imsmaildomain=mail_domain)['result']
            entry_attrs['imsmailboxclassdn'] = mail_domain_config.get('imsmailboxdefaultclassdn')[0]
        except Exception as e:
            logger.debug("mail_domain %s Error: %s", mail_domain, traceback.format_exc())

    if 'imsmailboxquota' in entry_attrs:
        try:
            quota = int(entry_attrs['imsmailboxquota'])
        except errors.NotFound:
            raise errors.ValidationError(name='imsmailboxquota', error='not a number')
        quota_str = '*:storage={}M'.format(quota)
        entry_attrs['imsmailboxquota'] = quota_str
    else:
        try:
            mail_domain_config = self.api.Command['maildomain_show'](all=True, raw=True, imsmaildomain=mail_domain)['result']
            quota = int(mail_domain_config.get('imsmaildefaultquota')[0])
            quota_str = '*:storage={}M'.format(quota)
            entry_attrs['imsmailboxquota'] = quota_str
        except Exception as e:
            logger.debug("mail_domain %s Error: %s", mail_domain, traceback.format_exc())

    entry_attrs['imsmailreceiveexternally'] = entry_attrs.get('imsmailreceiveexternally', True)
    entry_attrs['imsmailsendexternally'] = entry_attrs.get('imsmailsendexternally', True)

    return dn


user_add.register_pre_callback(useradd_pre_callback)


def useradd_post_callback(self, ldap, dn, entry_attrs, *keys, **options):
    if 'imsmailboxquota' in entry_attrs:
        try:
            _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxquota'][0]))
            entry_attrs['imsmailboxquota'] = str(limit)
        except ValueError:
            pass

    return dn


user_add.register_post_callback(useradd_post_callback)


def usermod_pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
    if 'objectclass' in entry_attrs:
        obj_classes = entry_attrs['objectclass']
    else:
        obj_classes = ldap.get_entry(dn, ['objectclass'])['objectclass']

    obj_classes = [o.lower() for o in obj_classes]
    mail_obj_classes = {'imsmailsender', 'imsmailrecipent'}

    if mail_obj_classes.intersection(obj_classes):
        normalize_mail_attrs(entry_attrs)

    if 'imsmailboxquota' in entry_attrs:
        quota_str = '*:storage={}M'.format(entry_attrs['imsmailboxquota'])
        entry_attrs['imsmailboxquota'] = quota_str

    return dn


user_mod.register_pre_callback(usermod_pre_callback)


def usermod_post_callback(self, ldap, dn, entry_attrs, *keys, **options):
    if 'imsmailboxquota' in entry_attrs:
        try:
            _, _, limit = parse_quota_rule(str(entry_attrs['imsmailboxquota'][0]))
            entry_attrs['imsmailboxquota'] = str(limit)
        except ValueError:
            pass

    return dn


user_mod.register_post_callback(usermod_post_callback)


@register()
class user_migrate_mail(LDAPUpdate):
    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        _entry_attrs = ldap.get_entry(dn, ['objectclass', 'mail'])
        obj_classes = entry_attrs['objectclass'] = _entry_attrs['objectclass']

        obj_classes = [o.lower() for o in obj_classes]
        mail_obj_classes = {'imsmailsender', 'imsmailrecipent', 'imsmailboxuser'}

        if mail_obj_classes.intersection(obj_classes):
            raise MailAlreadyMigratedError

        entry_attrs['objectclass'].extend(mail_obj_classes.difference(obj_classes))

        if 'mail' not in _entry_attrs:
            config = ldap.get_ipa_config()
            defaultdomain = config.get('ipadefaultemaildomain', [None])[0]
            _entry_attrs['mail'] = [ keys[0] + u'@' + defaultdomain ]
            logger.debug("mail: %s", _entry_attrs['mail'][0])

        normalize_mail_attrs(_entry_attrs)
        entry_attrs['mail'] = _entry_attrs['mail']
        entry_attrs['imsmailaddress'] = _entry_attrs['imsmailaddress']

        mail_domain = entry_attrs['imsmailaddress'] .split("@")[1]

        try:
            mail_domain_config = self.api.Command['maildomain_show'](all=True, raw=True, imsmaildomain=mail_domain)['result']
            entry_attrs['imsmailboxclassdn'] = mail_domain_config.get('imsmailboxdefaultclassdn')[0]
        except errors.NotFound:
            pass

        try:
            mail_domain_config = self.api.Command['maildomain_show'](all=True, raw=True, imsmaildomain=mail_domain)['result']
            quota = mail_domain_config.get('imsmaildefaultquota')[0]
            quota_str = '*:storage={}M'.format(int(quota))
            entry_attrs['imsmailboxquota'] = quota_str
        except errors.NotFound:
            pass
        except TypeError:
            pass
        entry_attrs['imsmailreceiveexternally'] = entry_attrs.get('imsmailreceiveexternally', True)
        entry_attrs['imsmailsendexternally'] = entry_attrs.get('imsmailsendexternally', True)

        logger.debug("_entry_attrs: {0}".format(_entry_attrs))
        logger.debug("entry_attrs: {0}".format(entry_attrs))

        return dn


group.takes_params += (
    Str('imsmailalias*',
        cli_name='alias',
        label=_('Mail alias')
        ),
)

group.managed_permissions = {**group.managed_permissions, **{
    'System: Read Group Mail Attributes': {
        'ipapermbindruletype': 'all',
        'ipapermright': {'read', 'search', 'compare'},
        'ipapermdefaultattr': {'imsmailalias'}
    },
    'System: Modify Group Mail Attributes': {
        'ipapermbindruletype': 'permission',
        'ipapermright': {'write', 'add', 'delete'},
        'ipapermdefaultattr': {'imsmailalias'}
    }
}}


def groupadd_pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
    if 'imsmailalias' in entry_attrs:
        add_missing_object_class(ldap, 'imsmailrecipentgroup', dn, entry_attrs, update=False)
    return dn


group_add.register_pre_callback(groupadd_pre_callback)


def groupmod_pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
    if 'imsmailalias' in entry_attrs:
        entry_attrs.update(add_missing_object_class(ldap, 'imsmailrecipentgroup', dn, update=False))
    return dn


group_mod.register_pre_callback(groupmod_pre_callback)


@register()
class group_enable_mail(LDAPQuery):
    __doc__ = _('Enable mail aliases for the group.')

    has_output = output.standard_value
    msg_summary = _('Enabled mail aliases for group "%(value)s"')

    takes_options = (
        Str('imsmailalias+',
            cli_name='alias',
            label=_('Mail aliases')
            )
    )

    def execute(self, *args, **kw):
        dn = self.obj.get_dn(*args, **kw)
        entry = self.obj.backend.get_entry(dn, ['objectclass'])

        if 'imsmailrecipentgroup' not in [o.lower() for o in entry['objectclass']]:
            entry['objectclass'].append('imsmailrecipentgroup')
        else:
            raise errors.AlreadyActive()

        entry['imsmailalias'] = list(kw['imsmailalias'])

        self.obj.backend.update_entry(entry)

        return dict(
            result=True,
            value=pkey_to_value(args[0], kw),
        )


@register()
class group_disable_mail(LDAPQuery):
    __doc__ = _('Disable mail aliases for the group.')

    has_output = output.standard_value
    msg_summary = _('Disabled mail aliases for group "%(value)s"')

    def execute(self, *args, **kw):
        dn = self.obj.get_dn(*args, **kw)
        entry = self.obj.backend.get_entry(dn, ['objectclass', 'imsmailalias'])

        if 'imsmailrecipentgroup' in [o.lower() for o in entry['objectclass']]:
            try:
                del entry['imsmailalias']
            except KeyError:
                pass

            entry['objectclass'].remove('imsmailrecipentgroup')
        else:
            raise errors.AlreadyInactive()

        self.obj.backend.update_entry(entry)

        return dict(
            result=True,
            value=pkey_to_value(args[0], kw),
        )


host.takes_params += (
    Flag('disablemail',
         cli_name='disable_mail',
         label='Disable mail sending for host',
         flags=('no_search', 'virtual_attribute'),
         default=False
         ),
    Str('imsmailaddress?',
        cli_name='primary_mail',
        label=_('Primary mail address')
        ),
    Str('imsmailsenderalias*',
        cli_name='send_alias',
        label=_('Allowed sender aliases')
        ),
    Bool('imsmailsendexternally?',
         cli_name='can_send_externally',
         label='Can send mails to external locations'
         ),
)

host.managed_permissions = {**host.managed_permissions, **{
    'System: Read Host Mail Attributes': {
        'ipapermbindruletype': 'all',
        'ipapermright': {'read', 'search', 'compare'},
        'ipapermdefaultattr': {'imsmailaddress', 'imsmailsenderalias', 'imsmailsendexternally'}
    },
    'System: Modify Host Mail Attributes': {
        'ipapermbindruletype': 'permission',
        'ipapermright': {'write', 'add', 'delete'},
        'ipapermdefaultattr': {'imsmailaddress', 'imsmailsenderalias', 'imsmailsendexternally'}
    }
}}


def hostadd_pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
    if not options.get('disablemail'):
        entry_attrs.update(add_missing_object_class(ldap, 'imsmailsender', dn, entry_attrs, update=False))
        if 'imsmailaddress' not in entry_attrs:
            config = ldap.get_ipa_config()
            entry_attrs['imsmailaddress'] = normalize_and_validate_email(entry_attrs['serverhostname'], config)

        entry_attrs['imsmailsendexternally'] = entry_attrs.get('imsmailsendexternally', False)

    return dn


host_add.register_pre_callback(hostadd_pre_callback)


@register()
class host_enable_mail(LDAPQuery):
    __doc__ = _('Enable mail sending for the host.')

    has_output = output.standard_value
    msg_summary = _('Enabled mail sending for host "%(value)s"')

    takes_options = (
        Str('imsmailaddress?',
            cli_name='primary_mail',
            label=_('Primary mail address')
            ),
        Str('imsmailsenderalias*',
            cli_name='send_alias',
            label=_('Allowed sender aliases')
            ),
        Flag('imsmailsendexternally',
             cli_name='can_send_externally',
             label='Can send mails to external locations',
             default=False),
    )

    def execute(self, *args, **kw):
        dn = self.obj.get_dn(*args, **kw)
        entry = self.obj.backend.get_entry(dn, ['objectclass', 'serverhostname'])

        if 'imsmailsender' not in [o.lower() for o in entry['objectclass']]:
            entry['objectclass'].append('imsmailsender')
        else:
            raise errors.AlreadyActive()

        if 'imsmailaddress' not in kw:
            config = self.obj.backend.get_ipa_config()
            entry['imsmailaddress'] = normalize_and_validate_email(entry['serverhostname'], config)
        else:
            entry['imsmailaddress'] = kw['imsmailaddress']

        if 'imsmailsenderalias' in kw:
            entry['imsmailsenderalias'] = list(kw['imsmailsenderalias'])

        entry['imsmailsendexternally'] = kw['imsmailsendexternally']

        self.obj.backend.update_entry(entry)

        return dict(
            result=True,
            value=pkey_to_value(args[0], kw),
        )


@register()
class host_disable_mail(LDAPQuery):
    __doc__ = _('Disable mail sending for the host.')

    has_output = output.standard_value
    msg_summary = _('Disabled mail sending for host "%(value)s"')

    def execute(self, *args, **kw):
        dn = self.obj.get_dn(*args, **kw)
        entry = self.obj.backend.get_entry(dn, ['objectclass', 'imsmailaddress', 'imsmailsendexternally'])

        if 'imsmailsender' in [o.lower() for o in entry['objectclass']]:
            try:
                del entry['imsmailaddress']
            except KeyError:
                pass

            try:
                del entry['imsmailsendexternally']
            except KeyError:
                pass

            while 'imsmailsender' in [o.lower() for o in entry['objectclass']]:
                entry['objectclass'].remove('imsmailsender')
        else:
            raise errors.AlreadyInactive()

        self.obj.backend.update_entry(entry)

        return dict(
            result=True,
            value=pkey_to_value(args[0], kw),
        )

